/* asciifig
 *
 * DESCRIPTION
 *  Draw figures using only plain ascii characters.
 *  Use the numeric keypad to move in 8 directions. 
 *  Moving combined with shift draws lines.
 *  The appropriate characters (-|/\...) are automatically chosen depending
 *  on wether the neigbouring characters are also 'on'.
 *
 * MOTIVATION
 *  The plain ascii format is often required, e.g for internet drafts, email...
 *  I don't know of any tool like this (emacs' picture-mode is the closest).  
 *  A good ascii back-end for nroff would also be nice.
 *
 * AUTHOR
 *  Wim Livens
 *
 * COPYRIGHT NOTICE
 *  Huh? what's that ?
 *  Do with it what ever you want but keep my name somewhere.
 *  I've wrote this both for fun and because I needed it.
 *  So if you make any improvements, I like to get them too!
 *
 * HISTORY
 *  1999-06-26 (late night)
 *   Only moving and drawing, |\/- chars are automatically chosen.
 *  1999-06-27 
 *   deleting, status line, check outside screen, save/load file
 *   cut/paste region, center figure, enter text.
 *  1999-06-28
 *   move mode instead of shift-keypad, that's less confortable but more 
 *   portable. I can't get the keybinding properly configured under X, so
 *   I assume users won't either, so I make my program very robust. It should 
 *   work at least with TERM=xterm. 
 *   
 *
 * TODO (sorted by priority)
 *  use attributes to distiguish lines and text
 *  real help window
 *  man-page
 *  refine ca array
 *  visualize region
 *  mouse support
 *  scrolling (horizontal & vertical)
 *  undo
 */

#include <ncurses.h>
#include <stdio.h>
#include <errno.h>

typedef struct region {
  int x1;
  int y1;
  int x2;
  int y2;
} region_t;


int x, y; /* cursor position */
int mx, my; /* mark position */
int mode;
bool saved;
char filename[128];
struct clipb {
  int x;
  int y;
  char *b;
} clipboard;


#define M_MOVE  1
#define M_DRAW  2
#define M_ERASE 3


/* what ascii char to use depending on neighbours 
 * the neighbours form a bitmap as follows: 
 * 0 1 2
 * 3   4
 * 5 6 7
 *
 * All the 'intelligence' for which character to use 
 * is in this array.  Please change it if you wish.
 */
const char ca[] = {
  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '+', '\\','|', '+', '/', 'v', '+', '+',
    '-', '+', '+', '#', '+', '+', '+', '#',
    '-', '+', '+', '+', '+', '+', '#', '#',
    '-', '-', '+', '#', '-', '-', '#', '#',
  /*...  ...  ...  ...  ...  ...  ...  ... */

  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '/', '>', '+', '+', '/', '+', '+', '+',
    '+', '+', '+', '#', '+', '+', '+', '#',
    '+', '+', '+', '+', '+', '+', '#', '#',
    '-', '-', '+', '#', '-', '+', '#', '#',
  /*+..  +..  +..  +..  +..  +..  +..  +.. */

  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '|', '+', '|', '|', '+', '+', '|', '|',
    '+', '+', '+', '#', '+', '+', '+', '#',
    '+', '+', '+', '+', '+', '+', '#', '#',
    '+', '+', '+', '#', '+', '+', '#', '#',
  /*.+.  .+.  .+.  .+.  .+.  .+.  .+.  .+. */

  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '+', '+', '|', '|', '+', '+', '|', '+',
    '#', '#', '#', '#', '#', '#', '#', '#',
    '+', '+', '+', '+', '+', '+', '#', '#',
    '#', '#', '#', '#', '#', '#', '#', '#',
  /*++.  ++.  ++.  ++.  ++.  ++.  ++.  ++. */

  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '\\','\\','+', '+', '<', '+', '+', '+',
    '+', '+', '+', '#', '+', '+', '+', '#',
    '+', '+', '+', '+', '+', '+', '#', '#',
    '-', '-', '+', '#', '-', '+', '#', '#',
  /*..+  ..+  ..+  ..+  ..+  ..+  ..+  ..+ */

  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '^', '+', '+', '+', '+', 'X', '+', '+',
    '+', '+', '+', '#', '+', '+', '+', '#',
    '+', '+', '+', '+', '+', '+', '#', '#',
    '-', '+', '+', '#', '+', '+', '#', '#',
  /*+.+  +.+  +.+  +.+  +.+  +.+  +.+  +.+ */

  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '+', '+', '|', '|', '+', '+', '|', '+',
    '+', '+', '+', '#', '+', '+', '+', '#',
    '#', '#', '#', '#', '#', '#', '#', '#',
    '#', '#', '#', '#', '#', '#', '#', '#',
  /*.++  .++  .++  .++  .++  .++  .++  .++ */

  /*...  +..  .+.  ++.  ..+  +.+  .++  +++ */
    '+', '+', '|', '+', '+', '+', '+', '+',
    '#', '#', '#', '#', '#', '#', '#', '#',
    '#', '#', '#', '#', '#', '#', '#', '#',
    '#', '#', '#', '#', '#', '#', '#', '#'
  /*+++  +++  +++  +++  +++  +++  +++  +++ */
};

/* test if x,y is outside the allowed region */
bool outside(int y, int x) {
  return (x<0 || x>=COLS || y<0 || y>=LINES-2);
}

/* test if character is 'on', i.e. not blank and not text */
bool ison(int y, int x) {
  char ch;
  ch=mvinch(y,x);
  return (ch=='+'||ch=='|'||ch=='-'||ch=='\\'||ch=='/'||
	  ch=='#'||ch=='V'||ch=='^'||ch=='<'||ch=='>');
}

/* calculate which charater to use depending on neigbours */
char calcchar(int y, int x) {
  int bitmap=0;
  if (!outside(y-1,x-1) && ison(y-1,x-1)) bitmap|=1;
  if (!outside(y-1,x  ) && ison(y-1,x  )) bitmap|=2;
  if (!outside(y-1,x+1) && ison(y-1,x+1)) bitmap|=4;
  if (!outside(y  ,x-1) && ison(y  ,x-1)) bitmap|=8;
  if (!outside(y  ,x+1) && ison(y  ,x+1)) bitmap|=16;
  if (!outside(y+1,x-1) && ison(y+1,x-1)) bitmap|=32;
  if (!outside(y+1,x  ) && ison(y+1,x  )) bitmap|=64;
  if (!outside(y+1,x+1) && ison(y+1,x+1)) bitmap|=128;
  return (ca[bitmap]);
}

/* draw status line */
void drawstatus() {
#define S_FILE_LEN 20  
  char *s, fileS[S_FILE_LEN], savedS[3], modeS[6] ;
  int i;

  s=malloc(COLS);

  if (saved) strcpy(savedS,"--"); else strcpy(savedS,"**");

  switch (mode) {
  case M_MOVE:  {strcpy(modeS,"MOVE-"); break;}
  case M_DRAW:  {strcpy(modeS,"DRAW-"); break;}
  case M_ERASE: {strcpy(modeS,"ERASE"); break;}
  default: { endwin(); fprintf(stderr, "asciifig: internal error: invalid mode %d\n", mode); exit(1);}
  }

  strncpy(fileS, filename, 20); 
  if (fileS[S_FILE_LEN-1]!=0) {
    fileS[S_FILE_LEN-2]='*';
    fileS[S_FILE_LEN-1]=0;
  } 
  /* pad filename with blanks
    else {
    for(i=strlen(fileS); i<S_FILE_LEN; i++)
      fileS[i]=' ';
  }
  */
  attron(A_REVERSE);
  sprintf(s,"---%s--%s--%s-(%3d,%2d)--asciifig-v0.1--? for help", savedS, fileS, modeS, x, y);
  mvaddstr(LINES-2,0,s);

  /* fill line with dashes */
  for (i=strlen(s); i<COLS; i++) {
    mvaddch(LINES-2,i,'-');
  }
  attroff(A_REVERSE);
  free(s);
}

void clearcmdline() {
  int i;
  for (i=0; i<COLS; i++) {
    mvaddch(LINES-1,i,' ');
  }

}

/* determine minimal region that contains the complete figure */
region_t clip() {
  int x,y;
  region_t r;
  r.x1=COLS; r.y1=LINES-2; r.x2=0; r.y2=0;
  for (y=0; y<LINES-2; y++)
    for (x=0; x<COLS; x++) {
      if (mvinch(y,x)!=' ') {
	if (r.x1>x) r.x1=x;
	if (r.y1>y) r.y1=y;
	if (r.x2<x) r.x2=x;
	if (r.y2<y) r.y2=y;
      }
    };
  return(r);
}

/* copy region to clipboard, clear region */
void cut(region_t r) {
  int x,y;

  clipboard.x=r.x2-r.x1+1;
  clipboard.y=r.y2-r.y1+1;

  for (y=0; y<clipboard.y; y++) {
    for (x=0; x<clipboard.x; x++) {
      *(clipboard.b + (x + clipboard.x * y))=mvinch(r.y1+y,r.x1+x);
      mvaddch(r.y1+y,r.x1+x,' ');
    }
  }
}

/* copy clipboard to screen */
void paste(int y, int x) {
  int cx, cy;

  for (cy=0; cy<clipboard.y; cy++) {
    for (cx=0; cx<clipboard.x; cx++) {
      if (!outside(y+cy, x+cx))
	mvaddch(y+cy, x+cx, *(clipboard.b + cx + (clipboard.x * cy)));
    }
  }
}
/* center figure on screen */
void center() {
  region_t r;
  r=clip();
  cut(r);
  paste((LINES-2-(r.y2-r.y1))/2,(COLS-(r.x2-r.x1))/2);
}

/* save screen contents to file */
void savefile() {
  int x,y;
  char s[S_FILE_LEN];
  char *sp;
  FILE *f;
  region_t r;

  if (saved) {
    clearcmdline();
    mvaddstr(LINES-1, 0, "(No changes need to be saved)");  
  } 
  else {
    clearcmdline();
    mvaddstr(LINES-1, 0, "filename: ");  
    
    /* use filename as default */
    for(sp=(char *)filename; *sp!=0; sp++);
    for(; sp!=filename; sp--) {
      if (*sp!=0) ungetch(*sp);
    }
    if (*sp!=0) ungetch(*sp);
    /* input */
    echo();
    wgetnstr(stdscr, filename, S_FILE_LEN);
    noecho();

    if (strcmp(filename,"")!=0) {    
      /* get minimal region surrounding figure */
      r=clip();

      /* write clipped region to file */
      f=fopen(filename, "w");
      if (f==NULL) {
	endwin(); 
	fprintf(stderr, "asciifig: error opening file %s: %s\n", filename,strerror(errno));
	exit(1);
      }
      /*    fprintf(f,"(%d,%d)-(%d,%d)\n",minx,miny, maxx,maxy);*/
      for (y=r.y1; y<=r.y2; y++) {
	for (x=r.x1; x<=r.x2; x++) {
	  fputc(mvinch(y,x),f);
	}
	fprintf(f,"\n");
      }
      fclose(f);
      
      clearcmdline();
      sprintf(s, "saved: %s", filename);
      mvaddstr(LINES-1, 0, s);
      saved=TRUE;
      drawstatus();
      refresh();
    } 
    else {
      sprintf(s, "not saved.");
      mvaddstr(LINES-1, 0, s);    
    }   
  }
}

void loadfile() {
  FILE *f;
  char ch;
  region_t r;

  /* read figure from file and write it in top-left corner */
  move(0,0);
  f=fopen(filename, "r");
  if (f==NULL) {
    endwin(); 
    fprintf(stderr, "asciifig: error opening file %s: %s\n", filename,strerror(errno));
    exit(1);
  }
  while (1) {
    ch=fgetc(f);
    if (ch==EOF) break;
    addch(ch);
  }
  if (ferror(f)) {
    endwin(); 
    fprintf(stderr, "asciifig: error reading file %s: %s\n", filename,strerror(errno));
    exit(1);
  }
  fclose(f);

  center();
}

main (int argc, char **argv) {
  int oldx,oldy, ch, i,j;
  bool shift;
  char s[80];
  region_t r;

  //  clipboard.b=malloc((size_t)(COLS*LINES));
  clipboard.b=malloc((size_t) 2048);

  /* initialize ncurses according to manpage */
  initscr(); cbreak(); noecho();
  nonl(); intrflush(stdscr, FALSE); keypad(stdscr, TRUE);

  mode=M_MOVE;
  saved=TRUE;
  if (argc<2) strcpy(filename,"");
  else {
    strcpy(filename, argv[1]);
    loadfile();
  }

  mx=-1; my=-1; 
  clipboard.x=0; clipboard.y=0;

  /* start in the center */
  x=COLS/2;y=LINES/2; 
  drawstatus();
  move(y,x);

  while (1) {

    shift=FALSE;
    oldx=x; oldy=y;

    ch=getch();
    switch (ch) {
    case '?':      
      sprintf(s,"Keys: arrows/num_keypad  Quit Save  Move Draw Erase Input-text  Mark Cut Paste"); 
      mvaddstr(LINES-1,0,s);
      goto nomove;
    case 'e': 
    case 'E': // erase
      mode=M_ERASE; goto nomove;
    case 'd':
    case 'D': // draw
      mode=M_DRAW; goto nomove;
    case 'w':
    case 'W': // move
      mode=M_MOVE; goto nomove;
    case 's': // save
      savefile(); goto nomove;
    case 'q': // quit
      savefile(); endwin(); exit(0);
    case 'i': // input
      sprintf(s,"input text. press ENTER when done."); 
      mvaddstr(LINES-1,0,s);
      move(y,x);
      echo();getstr(s);noecho();
      clearcmdline();
      saved=FALSE;
      goto nomove;
    case 'm': // mark
      mx=x; my=y;
      sprintf(s,"mark set."); 
      mvaddstr(LINES-1,0,s);
      goto nomove;
    case 'c': // cut
    case 'k': // kill
      if (mx>=0) {
	if (mx<x) {r.x1=mx;r.x2=x;} else {r.x1=x;r.x2=mx;}
	if (my<y) {r.y1=my;r.y2=y;} else {r.y1=y;r.y2=my;}
	cut(r);
	mx=-1;my=-1;
	saved=FALSE;
      }
      else {
	sprintf(s,"mark not set."); 
	mvaddstr(LINES-1,0,s);
      }
      goto nomove;
    case 'p': // paste
    case 'y': // yank
      if (clipboard.x>0) {
	paste(y,x);
	saved=FALSE;
      }
      else {
	sprintf(s,"clipboard is empty."); 
	mvaddstr(LINES-1,0,s);
      }
      goto nomove;
    case 'l': // center
      center();
      goto nomove;
 /* arrow keys */
      //    case 120:
    case '8':
    case KEY_UP:
      y--; break;
      //    case 119:
    case '2':
    case KEY_DOWN:
      y++; break;
      //    case 116:
    case '4':
    case KEY_LEFT:
      x--; break;
      //    case 118:
    case '6':
    case KEY_RIGHT:
      x++; break;
    case '7':
    case KEY_A1:
    case KEY_HOME:
      y--; x--; break;
    case '9':
    case KEY_A3:
    case KEY_PPAGE:
      y--; x++; break;
    case '1':
    case KEY_C1:
    case KEY_END:
      y++; x--; break;
    case '3':
    case KEY_C3:
    case KEY_NPAGE:
      y++; x++; break;
    case '5':
      break;
    default:
      sprintf(s, "key not bound.");
      mvaddstr(LINES-1,0,s);
    }
    //    sprintf(s,"KEY=[%o]",ch);
    //mvaddstr(1,1,s);

    clearcmdline();

    if (outside(y,x)) {x=oldx; y=oldy;}

    if (mode!=M_MOVE) {
      mx=-1;my=-1;
      saved=FALSE;
      /* draw/erase new character (temporary as a *) */
      switch (mode) {
      case M_DRAW: mvaddch(oldy,oldx,'+'); break;
      case M_ERASE: mvaddch(oldy,oldx,' '); break;
      }
      /* change char to the correct one (|, \, /, -, ...), 
	 do this also for the neighbours */
      for (i=-1; i<=1; i++) 
	for (j=-1; j<=1; j++) {
	  if (!outside(oldy+j, oldx+j) && ison(oldy+j, oldx+i)) {
	    ch=calcchar(oldy+j,oldx+i);
	    mvaddch(oldy+j,oldx+i,ch);
	  }
	};
    }
  nomove:
    drawstatus();
    move(y,x);  
    refresh();
  }

  endwin();
}
